#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <aisl/aisl.h>
#include "list.h"
#include "stream.h"
#include "client.h"

#define FLAG_KEEPALIVE (1<<0)
#define FLAG_HANDSHAKE (1<<1)
#define FLAG_CAN_READ  (1<<2)
#define FLAG_CAN_WRITE (1<<3)


struct aisl_client
{
  struct sockaddr_in   address;   /**< Client's address structure */
  aisl_server_t        server;    /**< Server instance associated with client */
  list_t               streams;   /**< list of client streams */
  struct buffer        in;        /**< Client's input buffer */
  struct buffer        out;       /**< Client's output buffer */
  #ifndef AISL_WITHOUT_SSL
  SSL                * ssl;       /**< SSL pointer for encrypted connections */
  #endif
  time_t               timestamp; /**< Last communication timestamp */

  aisl_stream_t        stream;    /**< Pending client's stream */
  /* int                  next_id;   / **< stream id generator (even)*/
  /* int                  istream;   / **< input stream id */
  /* int                  ostream;   / **< output stream id */
  int                  flags;     /**< Client's flag bitmask */
  int                  fd;        /**< Client's socket descriptor */

  aisl_http_version_t  protocol;  /**< Client's protocol version */
};


static bool
aisl_client_raise_event(aisl_client_t client, aisl_event_t event, ...)
{
  bool result;
  va_list vl_args;

  aisl_t instance = aisl_server_get_instance(client->server);

  va_start(vl_args, event);
  result = aisl_raise_vl(instance, client, event, vl_args);
  va_end(vl_args);

  return result;
}


static aisl_status_t
aisl_client_parse(aisl_client_t client, char * data, size_t size)
{
/* parse next data chunk */
    while ( p_status == PARSER_PENDING )
    {
      switch(s->state)
      {
        case STREAM_REQUEST_METHOD:
          p_status = parse_request_method(client, &ptr, &l);
          break;
        case STREAM_REQUEST_PATH:
          p_status = parse_request_path(client, &ptr, &l);
          break;
        case STREAM_REQUEST_PROTOCOL:
          p_status = parse_request_protocol(client, &ptr, &l);
          break;

        case STREAM_REQUEST_HEADER_KEY:
          p_status = parse_request_header_key(client, &ptr, &l);
          break;
        case STREAM_REQUEST_HEADER_VALUE:
          p_status = parse_request_header_value(client, &ptr, &l);
          break;

        case STREAM_REQUEST_CONTENT:
          p_status = parse_request_content(client, &ptr, &l);
          break;

        default:
          p_status = PARSER_FINISHED;
            /* this is error actually */

      }
    }

    if (p_status == PARSER_FAILED)
    {
      /* reply Bad Request here */
      client_close(client);
    }
    else if (l)
    {
      buffer_shift(s->buffer, s->buffer->size-l); /* reset buffer */
    }


}


/* In HTTP 2.0 client->stream will be NULL if stream related data was completely
 * parsed. If it is not NULL, then stream expects additional data -> same like
 * in mono stream HTTP 1.0 
 */
static aisl_status_t
aisl_client_input(aisl_client_t client)
{
  int                l;
  parser_status_t    p_status = PARSER_PENDING;
  aisl_stream_t      s = client->stream;

  char             * data = &client->in.data[ client->in.length ];
  size_t             size = client->in.size - client.in.length;

  #ifndef AISL_WITHOUT_SSL
  if (client->ssl)
  {
    if ( !(client->flags & FLAG_HANDSHAKE) )
    {
      if ( (l = SSL_accept(client->ssl)) != 1 )
      {
        l = SSL_get_error(client->ssl, l);

        if (l == SSL_ERROR_WANT_READ || l == SSL_ERROR_WANT_WRITE)
          return AISL_IDLE;

        fprintf( stderr
               , "*AISL: SSL handshake fail: %s\n"
               , ERR_error_string(l, NULL) );

        aisl_client_close(client);
        return AISL_SYSCALL_ERROR;
      }

      client->flags |= FLAG_HANDSHAKE;
    }

    l = SSL_read(client->ssl, data, size) ;
  }
  else
  #endif
    l = recv( client->fd, data, size, 0);

  if (l > 0)
  {
    data = client->in.data;
    size = client->in.length + l;

    client->in.length = size;

    return aisl_client_parse(client, data, size);
  }
  else if (l<0)
  {
    if (client->ssl)
    {
      if (SSL_get_error(client->ssl, l) == SSL_ERROR_WANT_READ)
        return AISL_IDLE;
    }
    else
    {
      if(errno == EWOULDBLOCK)
        return AISL_IDLE;
    }
  }

  /* both: client disconnect + on read error  */
  /* todo: raise client error here */
  client_close(client);

  return AISL_SYSCALL_ERROR;
}


static bool
aisl_client_output(aisl_client_t client)
{
  if (client->fd < 0)
  {
    fprintf(stderr, "[aisl] assertion !(client->fd<0) failed\n");
    return true;
  }

  stream_t s;
  int      l;

  s = list_index(client->streams, client->ostream);

  /*
  if (!s->c_length_unknown && s->buffer && s->buffer->len)
    buffer_move(gBuffer, s->buffer);
  */

  /* while stream is not flushed, we should raise event */
  if(s->flags & STREAM_FLAG_OUTPUT_CHUNKED)
  {
    /* in case of chunked output ( subscription for AISL_STREAM_OUTPUT event )
     * stream buffer will be initialized with OUTPUT_BUFFER_SIZE size, but
     * buffer->size will be used to carry amount of stored bytes
     * */
    size_t bsz = s->buffer->size;

    if (bsz < OUTPUT_BUFFER_SIZE)
    {
      if (buffer_clear(s->buffer, OUTPUT_BUFFER_SIZE) == 0)
        return false;

      s->buffer->size = bsz;
      bsz = OUTPUT_BUFFER_SIZE;
    }

    if ( (l = bsz - s->buffer->size) > OUTPUT_BUFFER_SIZE / 2 )
      aisl_raise_event( client->server->owner, s, AISL_STREAM_OUTPUT, l);
  }

  if (s->buffer->size == 0)
    return false;

  l = (client->ssl) ?
        SSL_write(client->ssl, s->buffer->data, s->buffer->size) :
        send(     client->fd,  s->buffer->data, s->buffer->size, 0);

  if (l > 0)
  {
    buffer_shift(s->buffer, l);
    if (s->state == STREAM_RESPONSE_READY && /* flushed */
        s->buffer->size == 0) /* all sent */
    {
      buffer_clear(s->buffer, 0);

      /* data has been sent */
      /*
      if (client->protocol == AISL_HTTP_2_0)
      {

      }
      else*/
      if (client->flags & CLIENT_FLAG_KEEPALIVE)
      {
        list_remove(client->streams, s);
        stream_free(s);

        s = stream_new((struct sockaddr_in  *) client, client->next_id++, STREAM_REQUEST_METHOD );
        list_append(client->streams, s);
      }
      else
      {
        client_close(client);
      }
    }
    return true;
  }

  /* l < 0 */
  if (client->ssl)
  {
    if ( SSL_get_error(client->ssl, l) == SSL_ERROR_WANT_WRITE )
      return false;
  }
  else
  {
    if (errno == EWOULDBLOCK)
      return false;
  }

  client_close(client);

  return true;
}


aisl_client_t
aisl_client_new( aisl_server_t        server,
                 int                  fd,
                 struct sockaddr_in * addr,
                 SSL_CTX            * ssl_ctx )
{
  aisl_client_t   client;
  aisl_stream_t   stream;

  if ( (client = calloc(1, sizeof(struct client))) != NULL )
  {
    memcpy(&client->address, addr, sizeof(struct sockaddr_in));
    client->fd         = fd;
    client->next_id    = 2;
    client->protocol   = AISL_HTTP_1_0;
    client->timestamp  = time(NULL);
    client->flags      = FLAG_KEEPALIVE | FLAG_HANDSHAKE | FLAG_CAN_READ;

    if (buffer_init(client->in, 2*BUFFER_SIZE) == 0)
    {
      memcpy(&client->out, client->in, sizeof(struct buffer));

      stream = aisl_stream_new(client, 0, STREAM_REQUEST_METHOD);

      if (stream != NULL)
      {
        client->stream = stream;

        #ifdef AISL_WITHOUT_SSL

        return client;

        #else

        if ( !ssl_ctx )
          return client;

        if ((client->ssl = SSL_new(ssl_ctx)) != NULL )
        {
          SSL_set_fd(client->ssl, fd);
          return client;
        }

        #endif
      }
    }
    aisl_client_free(client);

  }

  return NULL;
}


void
aisl_client_free(aisl_client_t client)
{
  aisl_client_close(client);

  #ifndef AISL_WITHOUT_SSL
  if (client->ssl)
    SSL_free(client->ssl);
  #endif

  if (client->in.data)
    free(client->in.data);

  /* out buffer is a shared part of input buffer, so no need to free it */

  list_free(client->streams, (list_destructor_t)aisl_stream_free);

  free(client);
}


aisl_status_t
aisl_client_touch(aisl_client_t client)
{
  aisl_status_t result, status;
  aisl_stream_t s = client->stream;

  /* input */
  if (client->flags & FLAG_CAN_READ)
  {
    if ( (result = client_input(client)) < 0 )
      return result;
  }

  /* output */
  if (client->flags & FLAG_CAN_WRITE)
  {
    if ( (status = client_output(client)) < 0 )
      return status;
  }


  /*
  if ((client->protocol==AISL_HTTP_2_0 || s->state<STREAM_REQUEST_READY) &&
      (client_input(client)) ) result = true;
  */
  /* output */
  /*
  s = list_index(client->streams, client->ostream);

  if ( s->flags & (STREAM_FLAG_OUTPUT_READY | STREAM_FLAG_OUTPUT_CHUNKED) )
    result = client_output(client);
  */
  /* update timestamp */

  if (result == AISL_IDLE)
    result = status;

  if (result = AISL_SUCCESS)
    client->timestamp = time(NULL);

  return result;
}


bool
aisl_client_is_timed_out(aisl_client_t client, uint32_t timeout)
{
  time_t now;

  time(&now);

  if ( !(now - client->timestamp < timeout) )
  {
    aisl_client_raise_event( client, AISL_CLIENT_TIMEOUT );
    aisl_client_close(client);
    return true;
  }
  else
    return false;
}


int
aisl_client_get_socket(aisl_client_t client)
{
  return client->fd;
}


/* API Level ---------------------------------------------------------------- */

aisl_server_t
aisl_client_get_server(aisl_client_t client)
{
  return client->server;
}


bool
aisl_client_is_secure(aisl_client_t client)
{
  return (client->ssl == NULL) ? false : true;
}


bool
aisl_client_is_online(aisl_client_t client)
{
  return (client->fd == -1) ? false : true;
}


void
aisl_client_close(aisl_client_t client)
{
  if (client->fd != -1)
  {
    aisl_client_raise_event( client, AISL_CLIENT_DISCONNECT );

    close(client->fd);
    shutdown(client->fd, SHUT_RDWR);
    client->fd = -1;
  }
}


aisl_http_version_t
aisl_client_get_http_version(aisl_client_t client)
{
  return client->http_version;
}
